@@ -1,39 +1,29 @@
 #!/usr/bin/env python3
-# -*- coding: utf-8 -*-
-# vim: ts=4:et:sw=4:
 
-# ----------------------------------------------------------------------
-# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
-#
-# This program is Free Software and is released under the terms of
-#                    the GNU General License
-# ----------------------------------------------------------------------
+# --------------------------------------------------------------------
+# SPDX-License-Identifier: AGPL-3.0-or-later
+# © Copyright 2008-2024 José Manuel Rodríguez de la Rosa and contributors.
+# See the file CONTRIBUTORS.md for copyright details.
+# See https://www.gnu.org/licenses/agpl-3.0.html for details.
+# --------------------------------------------------------------------
 
-import collections
 import math
 import sys
 from math import pi as PI
 
 # typings
-from typing import List, NamedTuple, Optional
+from typing import NamedTuple
 
 import src.api.config
 import src.api.dataref
 import src.api.options
-import src.api.symboltable
-
-# Compiler API
-import src.api.symboltable.symboltable
 import src.api.utils
 
-# Lexers and parsers, etc
-import src.ply.yacc as yacc
-
 # Symbol Classes
 from src import arch
 
 # Global containers
-from src.api import errmsg
+from src.api import errmsg, opcodestemps
 from src.api import global_ as gl
 from src.api.check import (
     check_and_make_label,
@@ -45,6 +35,7 @@ from src.api.check import (
     is_number,
     is_numeric,
     is_static,
+    is_static_str,
     is_string,
     is_unsigned,
 )
@@ -53,7 +44,12 @@ from src.api.constants import CLASS, CONVENTION, SCOPE, LoopType
 from src.api.debug import __DEBUG__
 from src.api.errmsg import error, warning
 from src.api.global_ import LoopInfo
-from src.api.opcodestemps import OpcodesTemps
+
+# Compiler API
+from src.api.symboltable.symboltable import SymbolTable
+
+# Lexers and parsers, etc
+from src.ply import yacc
 from src.symbols import sym
 from src.symbols.id_ import SymbolID
 from src.symbols.symbol_ import Symbol
@@ -65,7 +61,7 @@ from src.zxbc.zxblex import tokens  # noqa
 # Function level entry ID in which scope we are into. If the list
 # is empty, we are at global scope
 # ----------------------------------------------------------------------
-FUNCTION_LEVEL: List[SymbolID] = gl.FUNCTION_LEVEL
+FUNCTION_LEVEL: list[SymbolID] = gl.FUNCTION_LEVEL
 
 # ----------------------------------------------------------------------
 # Function calls pending to check
@@ -81,7 +77,7 @@ INITS = gl.INITS
 # ----------------------------------------------------------------------
 # Global Symbol Table
 # ----------------------------------------------------------------------
-SYMBOL_TABLE = gl.SYMBOL_TABLE = src.api.symboltable.symboltable.SymbolTable()
+SYMBOL_TABLE = gl.SYMBOL_TABLE = SymbolTable()
 
 # ----------------------------------------------------------------------
 # Defined user labels. They all are prepended _label_. Line numbers 10,
@@ -92,17 +88,17 @@ LABELS = {}
 # ----------------------------------------------------------------------
 # True if we're in the middle of a LET sentence. False otherwise.
 # ----------------------------------------------------------------------
-LET_ASSIGNMENT = False
+LET_ASSIGNMENT: bool = False
 
 # ----------------------------------------------------------------------
 # True if PRINT sentence has been used.
 # ----------------------------------------------------------------------
-PRINT_IS_USED = False
+PRINT_IS_USED: bool = False
 
 # ----------------------------------------------------------------------
 # Last line number output for checking program key board BREAK
 # ----------------------------------------------------------------------
-last_brk_linenum = 0
+last_brk_linenum: int = 0
 
 
 # ----------------------------------------------------------------------
@@ -136,7 +132,8 @@ def init():
 
     ast = None
     data_ast = None  # Global Variables AST
-    optemps = OpcodesTemps()
+    opcodestemps.init()
+    optemps = opcodestemps.OpcodesTemps()
 
     gl.INITS.clear()
     del gl.FUNCTION_CALLS[:]
@@ -201,21 +198,21 @@ def make_number(value, lineno: int, type_=None):
     return sym.NUMBER(value, type_=type_, lineno=lineno)
 
 
-def make_typecast(type_: sym.TYPE, node: Optional[sym.SYMBOL], lineno: int):
+def make_typecast(type_: sym.TYPE, node: sym.SYMBOL | None, lineno: int) -> sym.TYPECAST | None:
     """Wrapper: returns a Typecast node"""
     if node is None or node.type_ is None:
-        return  # syntax / semantic error
+        return None  # syntax / semantic error
 
     assert isinstance(type_, sym.TYPE)
     return sym.TYPECAST.make_node(type_, node, lineno)
 
 
-def make_binary(lineno, operator, left, right, func=None, type_=None):
+def make_binary(lineno: int, operator, left, right, func=None, type_=None):
     """Wrapper: returns a Binary node"""
     return sym.BINARY.make_node(operator, left, right, lineno, func, type_)
 
 
-def make_unary(lineno, operator, operand, func=None, type_=None):
+def make_unary(lineno: int, operator, operand, func=None, type_=None):
     """Wrapper: returns a Unary node"""
     if operand is None:  # syntax / semantic error
         return None
@@ -223,18 +220,19 @@ def make_unary(lineno, operator, operand, func=None, type_=None):
     return sym.UNARY.make_node(lineno, operator, operand, func, type_)
 
 
-def make_builtin(lineno, fname, operands, func=None, type_=None):
+def make_builtin(lineno: int, fname: str, operands: Symbol | tuple | list | None, func=None, type_=None):
     """Wrapper: returns a Builtin function node.
     Can be a Symbol, tuple or list of Symbols
     If operand is an iterable, they will be expanded.
     """
     if operands is None:
         operands = []
-    assert isinstance(operands, Symbol) or isinstance(operands, tuple) or isinstance(operands, list)
-    # TODO: In the future, builtin functions will be implemented in an external library, like POINT or ATTR
-    __DEBUG__('Creating BUILTIN "{}"'.format(fname), 1)
-    if not isinstance(operands, collections.abc.Iterable):
+    assert isinstance(operands, Symbol | tuple | list)
+    # TODO: In the future, builtin functions will be implemented in an external stdlib, like POINT or ATTR
+    __DEBUG__(f'Creating BUILTIN "{fname}"', 1)
+    if not isinstance(operands, (list, tuple)):
         operands = [operands]
+
     return sym.BUILTIN.make_node(lineno, fname, func, type_, *operands)
 
 
@@ -242,7 +240,7 @@ def make_constexpr(lineno, expr):
     return sym.CONSTEXPR(expr, lineno=lineno)
 
 
-def make_strslice(lineno, s, lower, upper):
+def make_strslice(lineno: int, s, lower, upper):
     """Wrapper: returns String Slice node"""
     return sym.STRSLICE.make_node(lineno, s, lower, upper)
 
@@ -288,7 +286,7 @@ def make_arg_list(node, *args):
 def make_argument(expr, lineno: int, byref=None, name: str = None):
     """Wrapper: Creates a node containing an ARGUMENT"""
     if expr is None:
-        return  # There were a syntax / semantic error
+        return None  # There were a syntax / semantic error
 
     if byref is None:
         byref = OPTIONS.default_byref
@@ -315,7 +313,11 @@ def make_array_access(id_, lineno, arglist):
     This is an RVALUE (Read the element)
     """
     for i, arg in enumerate(arglist):
-        value = make_typecast(TYPE.by_name(src.api.constants.TYPE.to_string(gl.BOUND_TYPE)), arg.value, arg.lineno)
+        value = make_typecast(
+            TYPE.by_name(src.api.constants.TYPE.to_string(gl.BOUND_TYPE)),
+            arg.value,
+            arg.lineno,
+        )
         if value is None:  # semantic error?
             return None  # return error
         arg.value = value
@@ -323,7 +325,7 @@ def make_array_access(id_, lineno, arglist):
     return sym.ARRAYACCESS.make_node(id_, arglist, lineno, gl.FILENAME)
 
 
-def make_array_substr_assign(lineno: int, id_: str, arg_list, substr, expr_) -> Optional[sym.SENTENCE]:
+def make_array_substr_assign(lineno: int, id_: str, arg_list, substr, expr_) -> sym.SENTENCE | None:
     if arg_list is None or substr is None or expr_ is None:
         return None  # There were errors
 
@@ -366,10 +368,10 @@ def make_call(id_: str, lineno: int, args: sym.ARGLIST):
     A "call" is just an ID followed by a list of arguments.
     E.g. a(4)
     - a(4) can be a function call if 'a' is a function
-    - a(4) can be a string slice if a is a string variable: a$(4)
-    - a(4) can be an access to an array if a is an array
+    - a(4) can be a string slice if 'a' is a string variable: a$(4)
+    - a(4) can be an access to an array if 'a' is an array
 
-    This function will inspect the id_. If it is undeclared then
+    This function will inspect the id_. If it is undeclared, then
     id_ will be taken as a forwarded function.
     """
     if args is None:
@@ -394,17 +396,21 @@ def make_call(id_: str, lineno: int, args: sym.ARGLIST):
             arr.append_child(offset)
         return arr
 
-    if entry.class_ == CLASS.var:  # An already declared/used string var
+    if entry.class_ in (CLASS.var, CLASS.const):  # An already declared/used string var
         if len(args) > 1:
             errmsg.syntax_error_not_array_nor_func(lineno, id_)
             return None
 
-        entry = SYMBOL_TABLE.access_var(id_, lineno)
-        if entry is None:
-            return None
+        if entry.class_ == CLASS.var:
+            entry = SYMBOL_TABLE.access_var(id_, lineno)
+            if entry is None:
+                return None
 
         if len(args) == 1:
-            return sym.STRSLICE.make_node(lineno, entry, args[0].value, args[0].value)
+            if entry.class_ == CLASS.var:
+                return make_strslice(lineno, entry, args[0].value, args[0].value)
+            # it's a const
+            return make_strslice(lineno, sym.STRING(entry.value, lineno), args[0].value, args[0].value)
 
         mark_entry_as_accessed(entry)
         return entry
@@ -412,7 +418,7 @@ def make_call(id_: str, lineno: int, args: sym.ARGLIST):
     return make_func_call(id_, lineno, args)
 
 
-def make_param_decl(id_: str, lineno: int, typedef, is_array: bool, default_value: Optional[sym.SYMBOL] = None):
+def make_param_decl(id_: str, lineno: int, typedef, is_array: bool, default_value: sym.SYMBOL | None = None):
     """Wrapper that creates a param declaration"""
     return SYMBOL_TABLE.declare_param(id_, lineno, typedef, is_array, default_value)
 
@@ -699,20 +705,27 @@ def p_var_decl_ini(p):
         typedef = sym.TYPEREF(expr.type_, p.lexer.lineno, implicit=True)
 
     value = make_typecast(typedef, expr, p.lineno(4))
-    defval = value if is_static(expr) else None
+    defval = value if is_static(expr) and value.type_ != TYPE.string else None
 
     if keyword == "DIM":
         SYMBOL_TABLE.declare_variable(idlist[0].name, idlist[0].lineno, typedef, default_value=defval)
     else:
         # keyword == "CONST"
         if defval is None:
-            errmsg.syntax_error_not_constant(p.lineno(4))
-            return
+            if not is_static_str(value):
+                errmsg.syntax_error_not_constant(p.lineno(4))
+                return
+            defval = value
 
         SYMBOL_TABLE.declare_const(idlist[0].name, idlist[0].lineno, typedef, default_value=defval)
 
     if defval is None:  # Okay do a delayed initialization
-        p[0] = make_sentence(p.lineno(1), "LET", SYMBOL_TABLE.access_var(idlist[0].name, p.lineno(1)), value)
+        p[0] = make_sentence(
+            p.lineno(1),
+            "LET",
+            SYMBOL_TABLE.access_var(idlist[0].name, p.lineno(1)),
+            value,
+        )
 
 
 def p_singleid(p):
@@ -752,14 +765,16 @@ def p_arr_decl_attr(p):
         if expr.token == "UNARY" and expr.operator == "ADDRESS":  # Must be an ID
             if expr.operand.token == "ARRAYACCESS":
                 if expr.operand.offset is None:
-                    error(p.lineno(4), "Address is not constant. Only constant subscripts are allowed")
+                    error(
+                        p.lineno(4),
+                        "Address is not constant. Only constant subscripts are allowed",
+                    )
                     return
             else:
                 if expr.operand.token not in ("ID", "VAR", "LABEL"):
                     error(p.lineno(3), "Only addresses of identifiers are allowed")
                     return
-                else:
-                    expr.operand.has_address = True
+                expr.operand.has_address = True
 
     elif not is_static(expr):
         errmsg.syntax_error_address_must_be_constant(p.lineno(3))
@@ -796,16 +811,23 @@ def p_arr_decl_initialized(p):
             if not isinstance(remaining, list):
                 return True  # It's OK :-)
 
-            error(lineno, "Unexpected extra vector dimensions. It should be %i" % len(remaining))
+            error(
+                lineno,
+                "Unexpected extra vector dimensions. It should be %i" % len(remaining),
+            )
             return False
 
         if not isinstance(remaining, list):
-            error(lineno, "Mismatched vector size. Missing %i extra dimension(s)" % len(boundlist))
+            error(
+                lineno,
+                "Mismatched vector size. Missing %i extra dimension(s)" % len(boundlist),
+            )
             return False
 
         if len(remaining) != boundlist[0].count:
             error(
-                lineno, "Mismatched vector size. Expected %i elements, got %i." % (boundlist[0].count, len(remaining))
+                lineno,
+                "Mismatched vector size. Expected %i elements, got %i." % (boundlist[0].count, len(remaining)),
             )
             return False  # It's wrong. :-(
 
@@ -927,7 +949,10 @@ def p_statement_border(p):
 def p_statement_plot(p):
     """statement : PLOT expr COMMA expr"""
     p[0] = make_sentence(
-        p.lineno(1), "PLOT", make_typecast(TYPE.ubyte, p[2], p.lineno(3)), make_typecast(TYPE.ubyte, p[4], p.lineno(3))
+        p.lineno(1),
+        "PLOT",
+        make_typecast(TYPE.ubyte, p[2], p.lineno(3)),
+        make_typecast(TYPE.ubyte, p[4], p.lineno(3)),
     )
 
 
@@ -1048,7 +1073,7 @@ def p_statement_call(p):
         p[0] = None
     elif len(p) == 2:
         entry = SYMBOL_TABLE.get_entry(p[1])
-        if not entry or entry.class_ in (CLASS.label, CLASS.unknown):
+        if entry is not None and entry.class_ in (CLASS.label, CLASS.unknown):
             p[0] = make_label(p[1], p.lineno(1))
         else:
             p[0] = make_sub_call(p[1], p.lineno(1), make_arg_list(None))
@@ -1136,11 +1161,17 @@ def p_array_copy(p):
         return
 
     if larray.ref.count != rarray.ref.count:
-        warning(l1, "Arrays '%s' and '%s' don't have the same number of dimensions" % (larray.name, rarray.name))
+        warning(
+            l1,
+            "Arrays '%s' and '%s' don't have the same number of dimensions" % (larray.name, rarray.name),
+        )
     else:
         for b1, b2 in zip(larray.ref.bounds, rarray.ref.bounds):
             if b1.count != b2.count:
-                warning(l1, "Arrays '%s' and '%s' don't have the same dimensions" % (array_id1, array_id2))
+                warning(
+                    l1,
+                    "Arrays '%s' and '%s' don't have the same dimensions" % (array_id1, array_id2),
+                )
                 break
     # Array copy
     mark_entry_as_accessed(larray)
@@ -1167,7 +1198,8 @@ def p_arr_assignment(p):
 
     if entry.type_ == TYPE.string:
         variable = gl.SYMBOL_TABLE.access_array(id_, p.lineno(i))
-        if len(variable.ref.bounds) + 1 == len(arg_list):
+        # variable is an array. If it has 0 bounds means they are undefined (param byref)
+        if len(variable.ref.bounds) and len(variable.ref.bounds) + 1 == len(arg_list):
             ss = arg_list.children.pop().value
             p[0] = make_array_substr_assign(p.lineno(i), id_, arg_list, (ss, ss), expr)
             return
@@ -1249,8 +1281,16 @@ def p_substr_assignment(p):
         )
     else:
         substr = (
-            make_typecast(_TYPE(gl.STR_INDEX_TYPE), make_number(gl.MIN_STRSLICE_IDX, lineno=p.lineno(2)), p.lineno(2)),
-            make_typecast(_TYPE(gl.STR_INDEX_TYPE), make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(2)), p.lineno(2)),
+            make_typecast(
+                _TYPE(gl.STR_INDEX_TYPE),
+                make_number(gl.MIN_STRSLICE_IDX, lineno=p.lineno(2)),
+                p.lineno(2),
+            ),
+            make_typecast(
+                _TYPE(gl.STR_INDEX_TYPE),
+                make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(2)),
+                p.lineno(2),
+            ),
         )
 
     lineno = p.lineno(2)
@@ -1560,10 +1600,16 @@ def p_for_sentence_start(p):
             warning(p.lineno(5), "STEP value is 0 and FOR might loop forever")
 
         if p[4].value > p[6].value and p[7].value > 0:
-            warning(p.lineno(5), "FOR start value is greater than end. This FOR loop is useless")
+            warning(
+                p.lineno(5),
+                "FOR start value is greater than end. This FOR loop is useless",
+            )
 
         if p[4].value < p[6].value and p[7].value < 0:
-            warning(p.lineno(2), "FOR start value is lower than end. This FOR loop is useless")
+            warning(
+                p.lineno(2),
+                "FOR start value is lower than end. This FOR loop is useless",
+            )
 
     id_type = common_type(common_type(p[4], p[6]), p[7])
     variable = SYMBOL_TABLE.access_var(p[2], p.lineno(2), default_type=id_type)
@@ -1602,7 +1648,13 @@ def p_end(p):
 def p_error_raise(p):
     """statement : ERROR expr"""
     q = make_number(1, lineno=p.lineno(2))
-    r = make_binary(p.lineno(1), "MINUS", make_typecast(TYPE.ubyte, p[2], p.lineno(1)), q, lambda x, y: x - y)
+    r = make_binary(
+        p.lineno(1),
+        "MINUS",
+        make_typecast(TYPE.ubyte, p[2], p.lineno(1)),
+        q,
+        lambda x, y: x - y,
+    )
     p[0] = make_sentence(p.lineno(1), "ERROR", r)
 
 
@@ -1612,7 +1664,13 @@ def p_stop_raise(p):
     """
     q = make_number(9, lineno=p.lineno(1)) if len(p) == 2 else p[2]
     z = make_number(1, lineno=p.lineno(1))
-    r = make_binary(p.lineno(1), "MINUS", make_typecast(TYPE.ubyte, q, p.lineno(1)), z, lambda x, y: x - y)
+    r = make_binary(
+        p.lineno(1),
+        "MINUS",
+        make_typecast(TYPE.ubyte, q, p.lineno(1)),
+        z,
+        lambda x, y: x - y,
+    )
     p[0] = make_sentence(p.lineno(1), "STOP", r)
 
 
@@ -1681,13 +1739,18 @@ def p_data(p):
         p[0] = None
         return
 
+    if gl.FUNCTION_LEVEL:
+        errmsg.error(p.lineno(1), "DATA not allowed within Functions nor Subs")
+        p[0] = None
+        return
+
     for d in p[2].children:
         value = d.value
         if is_static(value):
             datas_.append(d)
             continue
 
-        new_lbl = "__DATA__FUNCPTR__{0}".format(len(gl.DATA_FUNCTIONS))
+        new_lbl = f"__DATA__FUNCPTR__{len(gl.DATA_FUNCTIONS)}"
         entry = make_func_declaration(new_lbl, p.lineno(1), type_=value.type_, class_=CLASS.function)
         if not entry:
             continue
@@ -1757,7 +1820,11 @@ def p_read(p):
 
         if isinstance(entry, sym.ARRAYLOAD):
             reads.append(
-                make_sentence(p.lineno(1), "READ", sym.ARRAYACCESS(entry.entry, entry.args, entry.lineno, gl.FILENAME))
+                make_sentence(
+                    p.lineno(1),
+                    "READ",
+                    sym.ARRAYACCESS(entry.entry, entry.args, entry.lineno, gl.FILENAME),
+                )
             )
             continue
 
@@ -1916,9 +1983,15 @@ def p_print_sentence(p):
     PRINT_IS_USED = True
 
 
+def p_print_elem_expr(p):
+    """print_elem : expr"""
+    p[0] = p[1]
+    if p[1] is not None and p[1].type_ == TYPE.boolean:
+        p[0] = make_typecast(TYPE.ubyte, p[1], p.lineno(1))
+
+
 def p_print_list_expr(p):
-    """print_elem : expr
-    | print_at
+    """print_elem : print_at
     | print_tab
     | attr
     | BOLD expr
@@ -2058,17 +2131,26 @@ def p_return_expr(p):
         return
 
     if is_numeric(p[2]) and FUNCTION_LEVEL[-1].type_ == TYPE.string:
-        error(p.lineno(2), "Type Error: Function must return a string, not a numeric value")
+        error(
+            p.lineno(2),
+            "Type Error: Function must return a string, not a numeric value",
+        )
         p[0] = None
         return
 
     if not is_numeric(p[2]) and FUNCTION_LEVEL[-1].type_ != TYPE.string:
-        error(p.lineno(2), "Type Error: Function must return a numeric value, not a string")
+        error(
+            p.lineno(2),
+            "Type Error: Function must return a numeric value, not a string",
+        )
         p[0] = None
         return
 
     p[0] = make_sentence(
-        p.lineno(1), "RETURN", FUNCTION_LEVEL[-1], make_typecast(FUNCTION_LEVEL[-1].type_, p[2], p.lineno(1))
+        p.lineno(1),
+        "RETURN",
+        FUNCTION_LEVEL[-1],
+        make_typecast(FUNCTION_LEVEL[-1].type_, p[2], p.lineno(1)),
     )
 
 
@@ -2161,11 +2243,10 @@ def p_save_code(p):
         if p[3].upper() not in ("SCREEN", "SCREEN$"):
             error(p.lineno(3), 'Unexpected "%s" ID. Expected "SCREEN$" instead' % p[3])
             return None
-        else:
-            # ZX Spectrum screen start + length
-            # This should be stored in a architecture-dependant file
-            start = make_number(16384, lineno=p.lineno(1))
-            length = make_number(6912, lineno=p.lineno(1))
+        # ZX Spectrum screen start + length
+        # This should be stored in a architecture-dependant file
+        start = make_number(16384, lineno=p.lineno(1))
+        length = make_number(6912, lineno=p.lineno(1))
     else:
         start = make_typecast(TYPE.uinteger, p[4], p.lineno(4))
         length = make_typecast(TYPE.uinteger, p[6], p.lineno(6))
@@ -2225,13 +2306,12 @@ def p_load_code(p):
         if p[3].upper() not in ("SCREEN", "SCREEN$", "CODE"):
             error(p.lineno(3), 'Unexpected "%s" ID. Expected "SCREEN$" instead' % p[3])
             return None
-        else:
-            if p[3].upper() == "CODE":  # LOAD "..." CODE
-                start = make_number(0, lineno=p.lineno(3))
-                length = make_number(0, lineno=p.lineno(3))
-            else:  # SCREEN$
-                start = make_number(16384, lineno=p.lineno(3))
-                length = make_number(6912, lineno=p.lineno(3))
+        if p[3].upper() == "CODE":  # LOAD "..." CODE
+            start = make_number(0, lineno=p.lineno(3))
+            length = make_number(0, lineno=p.lineno(3))
+        else:  # SCREEN$
+            start = make_number(16384, lineno=p.lineno(3))
+            length = make_number(6912, lineno=p.lineno(3))
     else:
         start = make_typecast(TYPE.uinteger, p[4], p.lineno(3))
 
@@ -2332,7 +2412,13 @@ def p_expr_shl_expr(p):
     if p[1].type_ in (TYPE.float_, TYPE.fixed):
         p[1] = make_typecast(TYPE.ulong, p[1], p.lineno(2))
 
-    p[0] = make_binary(p.lineno(2), "SHL", p[1], make_typecast(TYPE.ubyte, p[3], p.lineno(2)), lambda x, y: x << y)
+    p[0] = make_binary(
+        p.lineno(2),
+        "SHL",
+        p[1],
+        make_typecast(TYPE.ubyte, p[3], p.lineno(2)),
+        lambda x, y: x << y,
+    )
 
 
 def p_expr_shr_expr(p):
@@ -2344,7 +2430,13 @@ def p_expr_shr_expr(p):
     if p[1].type_ in (TYPE.float_, TYPE.fixed):
         p[1] = make_typecast(TYPE.ulong, p[1], p.lineno(2))
 
-    p[0] = make_binary(p.lineno(2), "SHR", p[1], make_typecast(TYPE.ubyte, p[3], p.lineno(2)), lambda x, y: x >> y)
+    p[0] = make_binary(
+        p.lineno(2),
+        "SHR",
+        p[1],
+        make_typecast(TYPE.ubyte, p[3], p.lineno(2)),
+        lambda x, y: x >> y,
+    )
 
 
 def p_minus_expr(p):
@@ -2474,6 +2566,11 @@ def p_string_lp_expr_rp(p):
 
 def p_expr_id_substr(p):
     """string : ID substr"""
+    entry = SYMBOL_TABLE.get_entry(p[1])
+    if entry is not None and entry.type_ == TYPE.string and entry.token == "CONST":
+        p[0] = make_strslice(p.lineno(1), entry, p[2][0], p[2][1])
+        return
+
     entry = SYMBOL_TABLE.access_var(p[1], p.lineno(1), default_type=TYPE.string)
     p[0] = None
     if entry is None:
@@ -2491,7 +2588,10 @@ def p_string_substr(p):
 def p_string_expr_lp(p):
     """string : LP expr RP substr"""
     if p[2].type_ != TYPE.string:
-        error(p.lexer.lineno, "Expected a string type expression. " "Got %s type instead" % TYPE.to_string(p[2].type_))
+        error(
+            p.lexer.lineno,
+            "Expected a string type expression. Got %s type instead" % TYPE.to_string(p[2].type_),
+        )
         p[0] = None
     else:
         p[0] = make_strslice(p.lexer.lineno, p[2], p[4][0], p[4][1])
@@ -2499,7 +2599,10 @@ def p_string_expr_lp(p):
 
 def p_subind_str(p):
     """substr : LP expr TO expr RP"""
-    p[0] = (make_typecast(TYPE.uinteger, p[2], p.lineno(1)), make_typecast(TYPE.uinteger, p[4], p.lineno(3)))
+    p[0] = (
+        make_typecast(TYPE.uinteger, p[2], p.lineno(1)),
+        make_typecast(TYPE.uinteger, p[4], p.lineno(3)),
+    )
 
 
 def p_subind_strTO(p):
@@ -2514,7 +2617,11 @@ def p_subind_TOstr(p):
     """substr : LP expr TO RP"""
     p[0] = (
         make_typecast(TYPE.uinteger, p[2], p.lineno(1)),
-        make_typecast(TYPE.uinteger, make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(4)), lineno=p.lineno(4)),
+        make_typecast(
+            TYPE.uinteger,
+            make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(4)),
+            lineno=p.lineno(4),
+        ),
         p.lineno(3),
     )
 
@@ -2523,7 +2630,11 @@ def p_subind_TO(p):
     """substr : LP TO RP"""
     p[0] = (
         make_typecast(TYPE.uinteger, make_number(0, lineno=p.lineno(2)), p.lineno(1)),
-        make_typecast(TYPE.uinteger, make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(3)), p.lineno(2)),
+        make_typecast(
+            TYPE.uinteger,
+            make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(3)),
+            p.lineno(2),
+        ),
     )
 
 
@@ -2543,7 +2654,10 @@ def p_id_expr(p):
 
     if entry.class_ == CLASS.array:  # HINT: This should never happen now
         if not LET_ASSIGNMENT:
-            error(p.lineno(1), "Variable '%s' is an array and cannot be used in this context" % p[1])
+            error(
+                p.lineno(1),
+                "Variable '%s' is an array and cannot be used in this context" % p[1],
+            )
             p[0] = None
     elif entry.class_ == CLASS.function:  # Function call with 0 args
         p[0] = make_call(p[1], p.lineno(1), make_arg_list(None))
@@ -2591,7 +2705,7 @@ def p_idcall_expr(p):
     if p[0] is None:
         return
 
-    if p[0].token in ("STRSLICE", "ID", "STRING"):
+    if p[0].token in ("STRSLICE", "ID", "STRING") or p[0].token == "CONST" and p[0].type_ == TYPE.string:
         entry = SYMBOL_TABLE.access_call(p[1], p.lineno(1))
         mark_entry_as_accessed(entry)
         return
@@ -2650,7 +2764,10 @@ def p_let_arr_substr_in_args(p):
 
     id_ = p[i]
     arg_list = p[i + 2]
-    substr = (arg_list.children.pop().value, make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(i + 3)))
+    substr = (
+        arg_list.children.pop().value,
+        make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(i + 3)),
+    )
     expr_ = p[i + 6]
     p[0] = make_array_substr_assign(p.lineno(i), id_, arg_list, substr, expr_)
 
@@ -2677,7 +2794,10 @@ def p_let_arr_substr_in_args3(p):
 
     id_ = p[i]
     arg_list = p[i + 2]
-    substr = (make_number(0, lineno=p.lineno(i + 4)), make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(i + 3)))
+    substr = (
+        make_number(0, lineno=p.lineno(i + 4)),
+        make_number(gl.MAX_STRSLICE_IDX, lineno=p.lineno(i + 3)),
+    )
     expr_ = p[i + 7]
     p[0] = make_array_substr_assign(p.lineno(i), id_, arg_list, substr, expr_)
 
@@ -2860,7 +2980,8 @@ def p_function_header_pre(p):
         for a, b in zip(p1, p2):
             if a.name != b.name:
                 warning(
-                    lineno, "Parameter '%s' in function '%s' has been renamed to '%s'" % (a.name, p[0].name, b.name)
+                    lineno,
+                    "Parameter '%s' in function '%s' has been renamed to '%s'" % (a.name, p[0].name, b.name),
                 )
 
             if a.type_ != b.type_ or a.byref != b.byref:
@@ -2886,7 +3007,10 @@ def p_function_header_pre(p):
 def p_function_error(p):
     """function_declaration : function_header program_co END error"""
     p[0] = None
-    error(p.lineno(3), "Unexpected token 'END'. Expected 'END FUNCTION' or 'END SUB' instead.")
+    error(
+        p.lineno(3),
+        "Unexpected token 'END'. Expected 'END FUNCTION' or 'END SUB' instead.",
+    )
 
 
 def p_function_def(p):
@@ -3035,12 +3159,18 @@ def p_function_body(p):
     | END SUB
     """
     if not FUNCTION_LEVEL:
-        error(p.lineno(3), "Unexpected token 'END %s'. No Function or Sub has been defined." % p[2])
+        error(
+            p.lineno(3),
+            "Unexpected token 'END %s'. No Function or Sub has been defined." % p[2],
+        )
         p[0] = None
         return
 
     a = FUNCTION_LEVEL[-1].class_
-    if a not in (CLASS.sub, CLASS.function):  # This function/sub was not correctly declared, so exit now
+    if a not in (
+        CLASS.sub,
+        CLASS.function,
+    ):  # This function/sub was not correctly declared, so exit now
         p[0] = None
         return
 
@@ -3048,7 +3178,10 @@ def p_function_body(p):
     b = p[i].lower()
 
     if a != b:
-        error(p.lineno(i), "Unexpected token 'END %s'. Should be 'END %s'" % (b.upper(), a.upper()))
+        error(
+            p.lineno(i),
+            "Unexpected token 'END %s'. Should be 'END %s'" % (b.upper(), a.upper()),
+        )
         p[0] = None
     else:
         p[0] = make_block() if p[1] == "END" else p[1]
@@ -3141,7 +3274,12 @@ def p_expr_usr(p):
     if p[2].type_ == TYPE.string:
         p[0] = make_builtin(p.lineno(1), "USR_STR", p[2], type_=TYPE.uinteger)
     else:
-        p[0] = make_builtin(p.lineno(1), "USR", make_typecast(TYPE.uinteger, p[2], p.lineno(1)), type_=TYPE.uinteger)
+        p[0] = make_builtin(
+            p.lineno(1),
+            "USR",
+            make_typecast(TYPE.uinteger, p[2], p.lineno(1)),
+            type_=TYPE.uinteger,
+        )
 
 
 def p_expr_rnd(p):
@@ -3153,7 +3291,12 @@ def p_expr_rnd(p):
 
 def p_expr_peek(p):
     """bexpr : PEEK bexpr %prec UMINUS"""
-    p[0] = make_builtin(p.lineno(1), "PEEK", make_typecast(TYPE.uinteger, p[2], p.lineno(1)), type_=TYPE.ubyte)
+    p[0] = make_builtin(
+        p.lineno(1),
+        "PEEK",
+        make_typecast(TYPE.uinteger, p[2], p.lineno(1)),
+        type_=TYPE.ubyte,
+    )
 
 
 def p_expr_peektype_(p):
@@ -3163,7 +3306,12 @@ def p_expr_peektype_(p):
 
 def p_expr_in(p):
     """bexpr : IN bexpr %prec UMINUS"""
-    p[0] = make_builtin(p.lineno(1), "IN", make_typecast(TYPE.uinteger, p[2], p.lineno(1)), type_=TYPE.ubyte)
+    p[0] = make_builtin(
+        p.lineno(1),
+        "IN",
+        make_typecast(TYPE.uinteger, p[2], p.lineno(1)),
+        type_=TYPE.ubyte,
+    )
 
 
 def p_expr_lbound(p):
@@ -3204,7 +3352,10 @@ def p_expr_lbound_expr(p):
         p[0] = None
         return
 
-    if is_number(num) and entry.scope in (SCOPE.local, SCOPE.global_):  # Try constant propagation
+    if is_number(num) and entry.scope in (
+        SCOPE.local,
+        SCOPE.global_,
+    ):  # Try constant propagation
         val = num.value
         if val < 0 or val > len(entry.bounds):
             error(p.lineno(6), "Dimension out of range")
@@ -3260,7 +3411,12 @@ def p_str(p):
     if is_number(p[2]):  # A constant is converted to string directly
         p[0] = sym.STRING(str(p[2].value), p.lineno(1))
     else:
-        p[0] = make_builtin(p.lineno(1), "STR", make_typecast(TYPE.float_, p[2], p.lineno(1)), type_=TYPE.string)
+        p[0] = make_builtin(
+            p.lineno(1),
+            "STR",
+            make_typecast(TYPE.float_, p[2], p.lineno(1)),
+            type_=TYPE.string,
+        )
 
 
 def p_inkey(p):
@@ -3293,10 +3449,10 @@ def p_val(p):
 
     def val(s):
         try:
-            x = float(s)
+            x = float(eval(s, {}, {}))
         except:
             x = 0
-            warning(p.lineno(1), "Invalid string numeric constant '%s' evaluated as 0" % s)
+            warning(p.lineno(1), f"Invalid string numeric constant '{s}' evaluated as 0")
         return x
 
     if p[2].type_ != TYPE.string:
@@ -3328,7 +3484,7 @@ def p_code(p):
 
 def p_sgn(p):
     """bexpr : SGN bexpr %prec UMINUS"""
-    sgn = lambda x: x < 0 and -1 or x > 0 and 1 or 0  # noqa
+    sgn = lambda x: x < 0 and -1 or x > 0 and 1 or 0
 
     if p[2].type_ == TYPE.string:
         error(p.lineno(1), "Expected a numeric expression, got TYPE.string instead")
@@ -3360,6 +3516,7 @@ def p_expr_trig(p):
             "EXP": math.exp,
             "SQR": math.sqrt,
         }[p[1]],
+        type_=TYPE.float_,
     )
 
 
@@ -3416,7 +3573,7 @@ def p_error(p):
         if loop_info.type == LoopType.FOR:
             errmsg.syntax_error_for_without_next(loop_info.lineno)
         else:
-            errmsg.syntax_error_loop_not_closed(loop_info.lineno, loop_info.type)
+            errmsg.syntax_error_loop_not_closed(loop_info.lineno, loop_info.type.value)
     # If there were previous errors, stop here
     # since this end of file is due to previous errors
     if gl.has_errors:
@@ -3433,4 +3590,4 @@ parser = src.api.utils.get_or_create("zxbparser", lambda: yacc.yacc(debug=True))
 
 ast = None
 data_ast = None  # Global Variables AST
-optemps = OpcodesTemps()
+optemps = opcodestemps.OpcodesTemps()
